# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['functional', 'functional.test']

package_data = \
{'': ['*'], 'functional.test': ['data/*', 'data/tmp/.gitkeep']}

install_requires = \
['dill>=0.3.2,<0.4.0', 'pandas[all]>=1.0.3,<2.0.0', 'tabulate<=1.0.0']

setup_kwargs = {
    'name': 'pyfunctional',
    'version': '1.4.1',
    'description': 'Package for creating data pipelines with chain functional programming',
    'long_description': '# PyFunctional\n![Build Status](https://github.com/EntilZha/PyFunctional/workflows/Python%20package/badge.svg)\n[![Code Coverage](https://codecov.io/github/EntilZha/PyFunctional/coverage.svg?branch=master)](https://codecov.io/github/EntilZha/PyFunctional?branch=master)\n[![ReadTheDocs](https://readthedocs.org/projects/scalafunctional/badge/?version=latest)](http://docs.pyfunctional.org)\n[![PyPI version](https://badge.fury.io/py/PyFunctional.svg)](https://badge.fury.io/py/PyFunctional)\n\n## Features\n`PyFunctional` makes creating data pipelines easy by using chained functional operators. Here are a\nfew examples of what it can do:\n\n* Chained operators: `seq(1, 2, 3).map(lambda x: x * 2).reduce(lambda x, y: x + y)`\n* Expressive and feature complete API\n* Read and write `text`, `csv`, `json`, `jsonl`, `sqlite`, `gzip`, `bz2`, and `lzma/xz` files\n* Parallelize "embarrassingly parallel" operations like `map` easily\n* Complete documentation, rigorous unit test suite, 100% test coverage, and CI which provide\nrobustness\n\n`PyFunctional`\'s API takes inspiration from Scala collections, Apache Spark RDDs, and Microsoft\nLINQ.\n\n## Table of Contents\n\n1. [Installation](#installation)\n2. [Examples](#examples)\n    1. [Simple Example](#simple-example)\n    2. [Aggregates and Joins](#aggregates-and-joins)\n    3. [Reading and Writing SQLite3](#readingwriting-sqlite3)\n    4. [Data Interchange with Pandas](https://github.com/EntilZha/PyFunctional/blob/master/examples/PyFunctional-pandas-tutorial.ipynb)\n3. [Writing to Files](#writing-to-files)\n4. [Parallel Execution](#parallel-execution)\n5. [Github Shortform Documentation](#documentation)\n    1. [Streams, Transformations, and Actions](#streams-transformations-and-actions)\n    2. [Streams API](#streams-api)\n    3. [Transformations and Actions APIs](#transformations-and-actions-apis)\n    4. [Lazy Execution](#lazy-execution)\n6. [Contributing and Bug Fixes](#contributing-and-bug-fixes)\n7. [Changelog](https://github.com/EntilZha/PyFunctional/blob/master/CHANGELOG.md)\n\n## Installation\n`PyFunctional` is available on [pypi](https://pypi.python.org/pypi/PyFunctional) and can be\ninstalled by running:\n\n```bash\n# Install from command line\n$ pip install pyfunctional\n```\n\nThen in python run: `from functional import seq`\n\n## Examples\n`PyFunctional` is useful for many tasks, and can natively open several common file types. Here\nare a few examples of what you can do.\n\n### Simple Example\n```python\nfrom functional import seq\n\nseq(1, 2, 3, 4)\\\n    .map(lambda x: x * 2)\\\n    .filter(lambda x: x > 4)\\\n    .reduce(lambda x, y: x + y)\n# 14\n\n# or if you don\'t like backslash continuation\n(seq(1, 2, 3, 4)\n    .map(lambda x: x * 2)\n    .filter(lambda x: x > 4)\n    .reduce(lambda x, y: x + y)\n)\n# 14\n```\n\n### Streams, Transformations and Actions\n`PyFunctional` has three types of functions:\n\n1. Streams: read data for use by the collections API.\n2. Transformations: transform data from streams with functions such as `map`, `flat_map`, and\n`filter`\n3. Actions: These cause a series of transformations to evaluate to a concrete value. `to_list`,\n`reduce`, and `to_dict` are examples of actions.\n\nIn the expression `seq(1, 2, 3).map(lambda x: x * 2).reduce(lambda x, y: x + y)`, `seq` is the\nstream, `map` is the transformation, and `reduce` is the action.\n\n### Filtering a list of account transactions\n```python\nfrom functional import seq\nfrom collections import namedtuple\n\nTransaction = namedtuple(\'Transaction\', \'reason amount\')\ntransactions = [\n    Transaction(\'github\', 7),\n    Transaction(\'food\', 10),\n    Transaction(\'coffee\', 5),\n    Transaction(\'digitalocean\', 5),\n    Transaction(\'food\', 5),\n    Transaction(\'riotgames\', 25),\n    Transaction(\'food\', 10),\n    Transaction(\'amazon\', 200),\n    Transaction(\'paycheck\', -1000)\n]\n\n# Using the Scala/Spark inspired APIs\nfood_cost = seq(transactions)\\\n    .filter(lambda x: x.reason == \'food\')\\\n    .map(lambda x: x.amount).sum()\n\n# Using the LINQ inspired APIs\nfood_cost = seq(transactions)\\\n    .where(lambda x: x.reason == \'food\')\\\n    .select(lambda x: x.amount).sum()\n\n# Using PyFunctional with fn\nfrom fn import _\nfood_cost = seq(transactions).filter(_.reason == \'food\').map(_.amount).sum()\n```\n\n### Aggregates and Joins\nThe account transactions example could be done easily in pure python using list comprehensions. To\nshow some of the things `PyFunctional` excels at, take a look at a couple of word count examples.\n\n```python\nwords = \'I dont want to believe I want to know\'.split(\' \')\nseq(words).map(lambda word: (word, 1)).reduce_by_key(lambda x, y: x + y)\n# [(\'dont\', 1), (\'I\', 2), (\'to\', 2), (\'know\', 1), (\'want\', 2), (\'believe\', 1)]\n```\n\nIn the next example we have chat logs formatted in [json lines (jsonl)](http://jsonlines.org/) which\ncontain messages and metadata. A typical jsonl file will have one valid json on each line of a file.\nBelow are a few lines out of `examples/chat_logs.jsonl`.\n\n```json\n{"message":"hello anyone there?","date":"10/09","user":"bob"}\n{"message":"need some help with a program","date":"10/09","user":"bob"}\n{"message":"sure thing. What do you need help with?","date":"10/09","user":"dave"}\n```\n\n```python\nfrom operator import add\nimport re\nmessages = seq.jsonl(\'examples/chat_logs.jsonl\')\n\n# Split words on space and normalize before doing word count\ndef extract_words(message):\n    return re.sub(\'[^0-9a-z ]+\', \'\', message.lower()).split(\' \')\n\n\nword_counts = messages\\\n    .map(lambda log: extract_words(log[\'message\']))\\\n    .flatten().map(lambda word: (word, 1))\\\n    .reduce_by_key(add).order_by(lambda x: x[1])\n\n```\n\nNext, lets continue that example but introduce a json database of users from `examples/users.json`.\nIn the previous example we showed how `PyFunctional` can do word counts, in the next example lets\nshow how `PyFunctional` can join different data sources.\n\n```python\n# First read the json file\nusers = seq.json(\'examples/users.json\')\n#[(\'sarah\',{\'date_created\':\'08/08\',\'news_email\':True,\'email\':\'sarah@gmail.com\'}),...]\n\nemail_domains = users.map(lambda u: u[1][\'email\'].split(\'@\')[1]).distinct()\n# [\'yahoo.com\', \'python.org\', \'gmail.com\']\n\n# Join users with their messages\nmessage_tuples = messages.group_by(lambda m: m[\'user\'])\ndata = users.inner_join(message_tuples)\n# [(\'sarah\',\n#    (\n#      {\'date_created\':\'08/08\',\'news_email\':True,\'email\':\'sarah@gmail.com\'},\n#      [{\'date\':\'10/10\',\'message\':\'what is a...\',\'user\':\'sarah\'}...]\n#    )\n#  ),...]\n\n# From here you can imagine doing more complex analysis\n```\n\n### CSV, Aggregate Functions, and Set functions\nIn `examples/camping_purchases.csv` there are a list of camping purchases. Lets do some cost\nanalysis and compare it the required camping gear list stored in `examples/gear_list.txt`.\n\n```python\npurchases = seq.csv(\'examples/camping_purchases.csv\')\ntotal_cost = purchases.select(lambda row: int(row[2])).sum()\n# 1275\n\nmost_expensive_item = purchases.max_by(lambda row: int(row[2]))\n# [\'4\', \'sleeping bag\', \' 350\']\n\npurchased_list = purchases.select(lambda row: row[1])\ngear_list = seq.open(\'examples/gear_list.txt\').map(lambda row: row.strip())\nmissing_gear = gear_list.difference(purchased_list)\n# [\'water bottle\',\'gas\',\'toilet paper\',\'lighter\',\'spoons\',\'sleeping pad\',...]\n```\n\nIn addition to the aggregate functions shown above (`sum` and `max_by`) there are many more.\nSimilarly, there are several more set like functions in addition to `difference`.\n\n### Reading/Writing SQLite3\n`PyFunctional` can read and write to SQLite3 database files. In the example below, users are read\n from `examples/users.db` which stores them as rows with columns `id:Int` and `name:String`.\n\n```python\ndb_path = \'examples/users.db\'\nusers = seq.sqlite3(db_path, \'select * from user\').to_list()\n# [(1, \'Tom\'), (2, \'Jack\'), (3, \'Jane\'), (4, \'Stephan\')]]\n\nsorted_users = seq.sqlite3(db_path, \'select * from user order by name\').to_list()\n# [(2, \'Jack\'), (3, \'Jane\'), (4, \'Stephan\'), (1, \'Tom\')]\n```\n\nWriting to a SQLite3 database is similarly easy\n\n```python\nimport sqlite3\nfrom collections import namedtuple\n\nwith sqlite3.connect(\':memory:\') as conn:\n    conn.execute(\'CREATE TABLE user (id INT, name TEXT)\')\n    conn.commit()\n    User = namedtuple(\'User\', \'id name\')\n    \n    # Write using a specific query\n    seq([(1, \'pedro\'), (2, \'fritz\')]).to_sqlite3(conn, \'INSERT INTO user (id, name) VALUES (?, ?)\')\n    \n    # Write by inserting values positionally from a tuple/list into named table\n    seq([(3, \'sam\'), (4, \'stan\')]).to_sqlite3(conn, \'user\')\n    \n    # Write by inferring schema from namedtuple\n    seq([User(name=\'tom\', id=5), User(name=\'keiga\', id=6)]).to_sqlite3(conn, \'user\')\n    \n    # Write by inferring schema from dict\n    seq([dict(name=\'david\', id=7), dict(name=\'jordan\', id=8)]).to_sqlite3(conn, \'user\')\n    \n    # Read everything back to make sure it wrote correctly\n    print(list(conn.execute(\'SELECT * FROM user\')))\n    \n    # [(1, \'pedro\'), (2, \'fritz\'), (3, \'sam\'), (4, \'stan\'), (5, \'tom\'), (6, \'keiga\'), (7, \'david\'), (8, \'jordan\')]\n```\n\n## Writing to files\nJust as `PyFunctional` can read from `csv`, `json`, `jsonl`, `sqlite3`, and text files, it can\nalso write them. For complete API documentation see the collections API table or the official docs.\n\n### Compressed Files\n`PyFunctional` will auto-detect files compressed with `gzip`, `lzma/xz`, and `bz2`. This is done\nby examining the first several bytes of the file to determine if it is compressed so therefore\nrequires no code changes to work.\n\nTo write compressed files, every `to_` function has a parameter `compression` which can be set to\nthe default `None` for no compression, `gzip` or `gz` for gzip compression, `lzma` or `xz` for lzma\ncompression, and `bz2` for bz2 compression.\n\n### Parallel Execution\nThe only change required to enable parallelism is to import `from functional import pseq` instead of\n`from functional import seq` and use `pseq` where you would use `seq`. The following\noperations are run in parallel with more to be implemented in a future release:\n\n* `map`/`select`\n* `filter`/`filter_not`/`where`\n* `flat_map`\n\nParallelization uses python `multiprocessing` and squashes chains of embarrassingly parallel\noperations to reduce overhead costs. For example, a sequence of maps and filters would be executed\nall at once rather than in multiple loops using `multiprocessing`\n\n## Documentation\nShortform documentation is below and full documentation is at\n[docs.pyfunctional.org](http://docs.pyfunctional.org/en/latest/functional.html).\n\n### Streams API\nAll of `PyFunctional` streams can be accessed through the `seq` object. The primary way to create\na stream is by calling `seq` with an iterable. The `seq` callable is smart and is able to accept\nmultiple types of parameters as shown in the examples below.\n\n```python\n# Passing a list\nseq([1, 1, 2, 3]).to_set()\n# [1, 2, 3]\n\n# Passing direct arguments\nseq(1, 1, 2, 3).map(lambda x: x).to_list()\n# [1, 1, 2, 3]\n\n# Passing a single value\nseq(1).map(lambda x: -x).to_list()\n# [-1]\n```\n\n`seq` also provides entry to other streams as attribute functions as shown below.\n\n```python\n# number range\nseq.range(10)\n\n# text file\nseq.open(\'filepath\')\n\n# json file\nseq.json(\'filepath\')\n\n# jsonl file\nseq.jsonl(\'filepath\')\n\n# csv file\nseq.csv(\'filepath\')\nseq.csv_dict_reader(\'filepath\')\n\n# sqlite3 db and sql query\nseq.sqlite3(\'filepath\', \'select * from data\')\n```\n\nFor more information on the parameters that these functions can take, reference the\n[streams documentation](http://docs.pyfunctional.org/en/latest/functional.html#module-functional.streams)\n\n### Transformations and Actions APIs\nBelow is the complete list of functions which can be called on a stream object from `seq`. For\ncomplete documentation reference\n[transformation and actions API](http://docs.pyfunctional.org/en/latest/functional.html#module-functional.pipeline).\n\nFunction | Description | Type\n ------- | ----------- | ----\n`map(func)/select(func)` | Maps `func` onto elements of sequence | transformation\n`starmap(func)/smap(func)` | Apply `func` to sequence with `itertools.starmap` | transformation\n`filter(func)/where(func)` | Filters elements of sequence to only those where `func(element)` is `True` | transformation\n`filter_not(func)` | Filters elements of sequence to only those where `func(element)` is `False` | transformation\n`flatten()` | Flattens sequence of lists to a single sequence | transformation\n`flat_map(func)` | `func` must return an iterable. Maps `func` to each element, then merges the result to one flat sequence | transformation\n`group_by(func)` | Groups sequence into `(key, value)` pairs where `key=func(element)` and `value` is from the original sequence | transformation\n`group_by_key()` | Groups sequence of `(key, value)` pairs by `key` | transformation\n`reduce_by_key(func)` | Reduces list of `(key, value)` pairs using `func` | transformation\n`count_by_key()` | Counts occurrences of each `key` in list of `(key, value)` pairs | transformation\n`count_by_value()` | Counts occurrence of each value in a list | transformation\n`union(other)` | Union of unique elements in sequence and `other` | transformation\n`intersection(other)` | Intersection of unique elements in sequence and `other` | transformation\n`difference(other)` | New sequence with unique elements present in sequence but not in `other` | transformation\n`symmetric_difference(other)` | New sequence with unique elements present in sequence or `other`, but not both | transformation\n`distinct()` | Returns distinct elements of sequence. Elements must be hashable | transformation\n`distinct_by(func)` | Returns distinct elements of sequence using `func` as a key | transformation\n`drop(n)` | Drop the first `n` elements of the sequence | transformation\n`drop_right(n)` | Drop the last `n` elements of the sequence | transformation\n`drop_while(func)` | Drop elements while `func` evaluates to `True`, then returns the rest | transformation\n`take(n)` | Returns sequence of first `n` elements | transformation\n`take_while(func)` | Take elements while `func` evaluates to `True`, then drops the rest | transformation\n`init()` | Returns sequence without the last element | transformation\n`tail()` | Returns sequence without the first element | transformation\n`inits()` | Returns consecutive inits of sequence | transformation\n`tails()` | Returns consecutive tails of sequence | transformation\n`zip(other)` | Zips the sequence with `other` | transformation\n`zip_with_index(start=0)` | Zips the sequence with the index starting at `start` on the right side | transformation\n`enumerate(start=0)` | Zips the sequence with the index starting at `start` on the left side | transformation\n`cartesian(*iterables, repeat=1)` | Returns cartesian product from itertools.product | transformation\n`inner_join(other)` | Returns inner join of sequence with other. Must be a sequence of `(key, value)` pairs | transformation\n`outer_join(other)` | Returns outer join of sequence with other. Must be a sequence of `(key, value)` pairs | transformation\n`left_join(other)` | Returns left join of sequence with other. Must be a sequence of `(key, value)` pairs | transformation\n`right_join(other)` | Returns right join of sequence with other. Must be a sequence of `(key, value)` pairs | transformation\n`join(other, join_type=\'inner\')` | Returns join of sequence with other as specified by `join_type`. Must be a sequence of `(key, value)` pairs | transformation\n`partition(func)` | Partitions the sequence into elements which satisfy `func(element)` and those that don\'t | transformation\n`grouped(size)` | Partitions the elements into groups of size `size` | transformation\n`sorted(key=None, reverse=False)/order_by(func)` | Returns elements sorted according to python `sorted` | transformation\n`reverse()` | Returns the reversed sequence | transformation\n`slice(start, until)` | Sequence starting at `start` and including elements up to `until` | transformation\n`head()` / `first()` | Returns first element in sequence | action\n`head_option()` | Returns first element in sequence or `None` if its empty | action\n`last()` | Returns last element in sequence | action\n`last_option()` | Returns last element in sequence or `None` if its empty | action\n`len()` / `size()` | Returns length of sequence | action\n`count(func)` | Returns count of elements in sequence where `func(element)` is True | action\n`empty()` | Returns `True` if the sequence has zero length | action\n`non_empty()` | Returns `True` if sequence has non-zero length | action\n`all()` | Returns `True` if all elements in sequence are truthy | action\n`exists(func)` | Returns `True` if `func(element)` for any element in the sequence is `True` | action\n`for_all(func)` | Returns `True` if `func(element)` is `True` for all elements in the sequence | action\n`find(func)` | Returns the element that first evaluates `func(element)` to `True` | action\n`any()` | Returns `True` if any element in sequence is truthy | action\n`max()` | Returns maximal element in sequence | action\n`min()` | Returns minimal element in sequence | action\n`max_by(func)` | Returns element with maximal value `func(element)` | action\n`min_by(func)` | Returns element with minimal value `func(element)` | action\n`sum()/sum(projection)` | Returns the sum of elements possibly using a projection | action\n`product()/product(projection)` | Returns the product of elements possibly using a projection | action\n`average()/average(projection)` | Returns the average of elements possibly using a projection | action\n`aggregate(func)/aggregate(seed, func)/aggregate(seed, func, result_map)` | Aggregate using `func` starting with `seed` or first element of list then apply `result_map` to the result | action\n`fold_left(zero_value, func)` | Reduces element from left to right using `func` and initial value `zero_value` | action\n`fold_right(zero_value, func)` | Reduces element from right to left using `func` and initial value `zero_value` | action\n`make_string(separator)` | Returns string with `separator` between each `str(element)` | action\n`dict(default=None)` / `to_dict(default=None)` | Converts a sequence of `(Key, Value)` pairs to a `dictionary`. If `default` is not None, it must be a value or zero argument callable which will be used to create a `collections.defaultdict` | action\n`list()` / `to_list()` | Converts sequence to a list | action\n`set() / to_set()` | Converts sequence to a set | action\n`to_file(path)` | Saves the sequence to a file at path with each element on a newline | action\n`to_csv(path)` | Saves the sequence to a csv file at path with each element representing a row | action\n`to_jsonl(path)` | Saves the sequence to a jsonl file with each element being transformed to json and printed to a new line | action\n`to_json(path)` | Saves the sequence to a json file. The contents depend on if the json root is an array or dictionary | action\n`to_sqlite3(conn, tablename_or_query, *args, **kwargs)` | Save the sequence to a SQLite3 db. The target table must be created in advance. | action\n`to_pandas(columns=None)` | Converts the sequence to a pandas DataFrame | action\n`cache()` | Forces evaluation of sequence immediately and caches the result | action\n`for_each(func)` | Executes `func` on each element of the sequence | action\n\n### Lazy Execution\nWhenever possible, `PyFunctional` will compute lazily. This is accomplished by tracking the list\nof transformations that have been applied to the sequence and only evaluating them when an action is\ncalled. In `PyFunctional` this is called tracking lineage. This is also responsible for the\nability for `PyFunctional` to cache results of computation to prevent expensive re-computation.\nThis is predominantly done to preserve sensible behavior and used sparingly. For example, calling\n`size()` will cache the underlying sequence. If this was not done and the input was an iterator,\nthen further calls would operate on an expired iterator since it was used to compute the length.\nSimilarly, `repr` also caches since it is most often used during interactive sessions where its\nundesirable to keep recomputing the same value. Below are some examples of inspecting lineage.\n\n```python\ndef times_2(x):\n    print(x)\n    return 2 * x\nelements = seq(1, 1, 2, 3, 4).map(times_2).distinct()\nelements._lineage\n# Lineage: sequence -> map(times_2) -> distinct\n\nl_elements = elements.to_list()\n# Prints: 1\n# Prints: 1\n# Prints: 2\n# Prints: 3\n# Prints: 4\n\nelements._lineage\n# Lineage: sequence -> map(times_2) -> distinct -> cache\n\nl_elements = elements.to_list()\n# The cached result is returned so times_2 is not called and nothing is printed\n```\n\nFiles are given special treatment if opened through the `seq.open` and related APIs.\n`functional.util.ReusableFile` implements a wrapper around the standard python file to support\nmultiple iteration over a single file object while correctly handling iteration termination and\nfile closing.\n\n## Road Map Idea\n* SQL based query planner and interpreter\n* `_` lambda operator\n\n## Contributing and Bug Fixes\nAny contributions or bug reports are welcome. Thus far, there is a 100% acceptance rate for pull\nrequests and contributors have offered valuable feedback and critique on code. It is great to hear\nfrom users of the package, especially what it is used for, what works well, and what could be\nimproved.\n\nTo contribute, create a fork of `PyFunctional`, make your changes, then make sure that they pass.\nIn order to be merged, all pull requests must:\n\n* Pass all the unit tests\n* Pass all the pylint tests, or ignore warnings with explanation of why its correct to do so\n* Not significantly reduce covrage without a good reason [coveralls.io](coveralls.io/github/EntilZha/PyFunctional))\n* Edit the `CHANGELOG.md` file in the `Next Release` heading with changes\n\n## Contact\n[Gitter for chat](https://gitter.im/EntilZha/PyFunctional)\n\n## Supported Python Versions\n* `PyFunctional` 1.4 and above supports and is tested against Python 3.6, Python 3.7, and PyPy3\n* `PyFunctional` 1.4 and above does not support python 2.7\n* `PyFunctional` 1.4 and above works in Python 3.5, but is not tested against it\n* `PyFunctional` 1.4 and above partially works in 3.8, parallel processing currently has issues, but other feature work fine\n* `PyFunctional` 1.3 and below supports and was tested against Python 2.7, Python 3.5, Python 3.6, PyPy2, and PyPy3\n\n\n## Changelog\n[Changelog](https://github.com/EntilZha/PyFunctional/blob/master/CHANGELOG.md)\n\n## About me\nTo learn more about me (the author) visit my webpage at\n[pedro.ai](https://www.pedro.ai).\n\nI created `PyFunctional` while using Python extensivel, and finding that I missed the\nease of use for manipulating data that Spark RDDs and Scala collections have. The project takes the\nbest ideas from these APIs as well as LINQ to provide an easy way to manipulate data when using\nScala is not an option or PySpark is overkill.\n\n## Contributors\nThese people have generously contributed their time to improving `PyFunctional`\n\n* [versae](https://github.com/versae)\n* [adrian17](https://github.com/adrian17)\n* [lucidfrontier45](https://github.com/lucidfrontier45)\n* [Digenis](https://github.com/Digenis)\n* [ChuyuHsu](https://github.com/ChuyuHsu)\n* [jsemric](https://github.com/jsemric)\n',
    'author': 'Pedro Rodriguez',
    'author_email': 'me@pedro.ai',
    'maintainer': 'Pedro Rodriguez',
    'maintainer_email': 'me@pedro.ai',
    'url': 'https://github.com/EntilZha/PyFunctional',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)
